/* eslint-disable import/no-extraneous-dependencies,no-param-reassign, no-underscore-dangle */
const path = require('path')
const fs = require('fs')
const Compression = require('compression-webpack-plugin')
const critical = require('critical')
const glob = require('glob')
const PurgecssPlugin = require('purgecss-webpack-plugin')
const PrerenderSPAPlugin = require('prerender-spa-plugin')

const Renderer = PrerenderSPAPlugin.PuppeteerRenderer

const srcDir = path.resolve('./src')

class CustomExtractor {
  static extract(content) {
    return content.match(/[A-Za-z0-9-_:/]+/g) || []
  }
}

module.exports = (api, projectOptions) => {
  const { pluginOptions } = projectOptions
  const optimizeOptions = pluginOptions ? pluginOptions.optimize || {} : {}
  const { rendererOptions, ...prerenderOptions } = optimizeOptions.prerender || {}

  api.chainWebpack((config) => {
    if (process.env.NODE_ENV === 'production') {
      // https://markus.oberlehner.net/blog/how-to-pre-render-vue-powered-websites-with-webpack/
      // https://github.com/SolarLiner/vue-cli-plugin-prerender-spa/blob/develop/index.js
      const rendererConfig = {
        headless: true,
        maxConcurrentRoutes: 4,
        renderAfterDocumentEvent: 'x-app-rendered',
        ...rendererOptions,
      }

      const renderer = new Renderer(rendererConfig)

      renderer.preServer = (Prerenderer) => {
        if (projectOptions.baseUrl) {
          const prefix = projectOptions.baseUrl
          const server = Prerenderer._server._expressServer
          server.use((req, res, next) => {
            if (req.url.indexOf(prefix) === 0) {
              req.url = req.url.slice(prefix.length - 1)
            }
            next()
          })
        }

        if (projectOptions.pages) {
          const server = Prerenderer._server._expressServer
          server.get('*', (req, res, next) => {
            if (!path.extname(req.url)) {
              const filePath = api.resolve(`${projectOptions.outputDir}${req.url}${path.basename(req.url) ? '.html' : 'index.html'}`)
              fs.exists(filePath, exists => (exists ? res.sendFile(filePath) : next()))
              return
            }
            next()
          })
        }
      }

      const outputDir = api.resolve(projectOptions.outputDir)
      const staticDir = api.resolve(projectOptions.outputDir)

      config.plugin('pre-render').use(PrerenderSPAPlugin, [
        {
          staticDir,
          outputDir,
          assetsDir: api.resolve(
            path.join(projectOptions.outputDir, projectOptions.assetsDir),
          ),
          indexPath: api.resolve(
            path.join(projectOptions.outputDir, projectOptions.indexPath),
          ),
          routes: ['/'],
          renderer,
          async postProcess(renderedRoute) {
            const { route } = renderedRoute

            if (route[route.length - 1] !== '/' && path.extname(route) === '') {
              renderedRoute.outputPath = path.join(outputDir || staticDir, `${route}.html`)
            }

            renderedRoute.html = renderedRoute.html
              .replace(/<script (.*?)>/g, '<script $1 async>')
              .replace('id="app"', 'id="app" data-server-rendered="true"')

            renderedRoute.html = (await critical.generate({
              inline: true,
              html: renderedRoute.html,
              base: 'dist/',
              folder: 'dist/',
              width: 1300,
              height: 900,
            })).toString('utf8')

            return renderedRoute
          },
          ...prerenderOptions,
        },
      ])
    }
    // remove the prefetch plugin
    config.plugins.delete('prefetch')

    // use ~ instead of @ because @ for srcDir is just confusing when npm uses it.
    config.resolve.alias.set('~', srcDir)

    // Entrypoint and Asset size should not be greater than 250kb combined (gziped)
    config.performance
      .maxEntrypointSize(125000) // 125kb
      .maxAssetSize(125000) // 125kb
      .assetFilter(assetFilename => assetFilename.endsWith('.js.gz'))

    // Gzip assets
    config
      .plugin('compress')
      .use(Compression, [
        {
          exclude: /.*.html$/,
        },
      ])
      .end()

    config.plugin('purge-css')
      .use(PurgecssPlugin, [{
        paths: () => glob.sync(`${srcDir}/**/!(*.test)*.{html,vue,css,scss}`, { nodir: true }),
        whitelist: () => ['app', 'html', 'body', 'fa', 'svg', 'animated'],
        whitelistPatterns: () => [
          /[A-Za-z0-9_-]+__[A-Za-z0-9_-]+/,
          /^fa-.*/,
          /^svg-.*/,
          /^animated.*/,
          /^slide.*/,
        ],
        extractors: [
          {
            extractor: CustomExtractor,
            extensions: ['html', 'vue', 'css', 'scss'],
          },
        ],
      },
      ])

    // Enable graphql files
    config.module
      .rule('graphql')
      .test(/\.graphql|\.gql$/)
      .use('graphql-tag/loader')
      .loader('graphql-tag/loader')

    // FIXME: https://github.com/graphql/graphql-js/issues/1272
    config.module
      .rule('mjs')
      .test(/\.mjs$/)
      .include.add(/node_modules/)
      .end()
      .type('javascript/auto')

    config.resolve.extensions
      .add('.mjs')
      .add('.gql')
      .add('.graphql')
  })
}
